/**
 * Kontrastanalaysverktyg.
 *
 * Infogar en dialogruta genom att klona en HTML-mall och hanterar funktionerna
 * i den. Dialogrutan innehåller följande komponenter:
 *
 * Förgrund:        redigerbart textfält, färgspecifikation + pipett.
 * Bakgrund:        redigerbart textfält, färgspecifikation + pipett.
 * Opacitet:        redigerbart numeriskt fält, decimaltal.
 * Stor text:       kryssruta för text som är minst 18pt eller 14pt fet.
 * Effektiv:        visningsfält, färgspecifikation (hexadecimalnotation) för
 *                  beräknat färgvärde vid transparens.
 * Annat än text:   kryssruta för kontrast i annat än text.
 * (exempel)        visar exempeltext med angiven för- och bakgrundsfärg.
 * Kontrast:        visningsfält för beräknat kontrastförhållande.
 * (sparade färger) komponent för att spara och återanvända färger.
 *   Namn:          användarens beteckning för en sparad färg.
 *   Fg:            knapp för att spara förgrundsfärgen.
 *   Bg:            knapp för att spara bakgrundsfärgen.
 *   Rensa:         knapp för att radera alla sparade färger.
 *   (lista)        lista med sparade färger.
 *     Fg:          sätt sparad färg som förgrundsfärg.
 *     Bg:          sätt sparad färg som bakgrundsfärg.
 *     ✘ (radera):  radera en sparad färg.
 *
 * Observerbara objekt
 * -------------------
 * Förändringar i vissa komponenters värden uppdaterar automatiskt värden eller
 * tillstånd i andra komponenter. Sådana komponenter är observerbara (instanser
 * av klassen `Observable`) hos vilka beroende komponenter kan registrera
 * lyssnarfunktioner som anropas när värdet i den observerade komponenten
 * förändras.
 *
 * Effektiv färg
 * -------------
 * Fältet för effektiv färg används när man vill beräkna kontraster för
 * transparenta element i flera lager. Det visar det beräknade effektiva
 * färgvärdet utan transaparens och används så här:
 * 1. Fyll i det nedersta transparenta elementets bakgrundsfärg som förgrund och
 *    underliggande icke-transparenta elements bakgrundsfärg som bakgrund.
 * 2. Sätt den beräknade effektiva bakgrunden som bakgrund och fyll i det
 *    ovanpåliggande elementets bakgrund som förgrundsfärg.
 * 3. Repetera steg 2 för varje elementskikt.
 * 4. Fyll i det översta elementet förgrundsfärg som förgrund och den beräknade
 *    effektiva bakgrunden som bakgrund. Avläs det beräknade kontrastför-
 *    hållandet.
 *
 * Primära klasser
 * ---------------
 * ColourParser      Tolkar färgspecifikationer.
 * ColourStore       Användargränssnitt för sparade färger.
 * ContrastAnalyzer  Beräknar kontrastförhållanden.
 * ContrastDialog    Dialogrutans primära användargränssnitt.
 * PersistentStorage Sparar och återkallar färger persistent.
 *
 * Klassen `ContrastDialog` håller alla indatavärden, genom att observera de
 * olika inmatningsfälten och kryssrutorna. Klassen `ColourStore` håller de
 * färger användaren väljer att spara och `PersistentStorage` hanterar lagring
 * och hämtning i persistenslagret (cookies), efter användarens medgivande.
 *
 * Sekundära klasser
 * -----------------
 * Checkbox    Observerbar kryssruta.
 * ColourField Observerbart inmatningsfält för färgspecifikationer.
 * NamedColour En färgspecifikationer, eventuellt namnsatt av användaren.
 * NumberField Observerbart inmatningsfält för numeriska värden.
 * Observable  Basklass för observerbara objekt.
 * Observer    Basklass för observatörer.
 */
((doc) => {
    "use strict";


    /**
     * Emulering av multipelt arv i ECMAScript 2015.
     *
     * Gör det möjligt att låta en klass utöka en basklass och en eller flera
     * mixin-klasser.
     *
     * @param baseClass Basklass att utöka.
     * @param mixins    En eller flera mixin-klasser.
     *
     * @return Class
     */
    const compound = (baseClass, ...mixins) => {
        // Hjälpfunktion för att kopiera egenskaper, symboler och metoder från
        // en mixin-klass till en basklass.
        const copyProps = (target, source) => {
            Object.getOwnPropertyNames(source).
                concat(Object.getOwnPropertySymbols(source)).
                    forEach((prop) => {
                        if (!prop.match(/^(constructor|prototype|arguments|caller|name|bind|call|apply|toString|length)$/)) {
                            Object.defineProperty(
                                target,
                                prop,
                                Object.getOwnPropertyDescriptor(source, prop)
                            );
                        }
            });
        };

        // Skapa en klass som utökar basklassen.
        class newClass extends baseClass {
            constructor (...args) {
                super(...args);
                mixins.forEach((mixin) => copyProps(this, new mixin));
            }
        }

        // Kopiera mixin-klasserna till den nya klassen.
        mixins.forEach((mixin) => {
            copyProps(newClass.prototype, mixin.prototype);
            copyProps(newClass,           mixin);
        });

        return newClass;
    };


    /**
     * Ett objekt som andra objekt kan observera.
     *
     * Observatörer kan registrera lyssnarfunktioner. När det observerade
     * objektet uppdateras anropas alla registrerade lyssnarfunktioner.
     *
     * @see Observer
     */
    class Observable {

        /**
         * Konstruera en observerbar instans.
         */
        constructor () {
            this._callbacks = {};
        }

        /**
         * Registrera en observatör.
         *
         * @param {Observar} observer Observatör.
         * @param {callback} callback Funktion att anropa när det observerade
         *                            objektet ändrar värde.
         * @param {string}   event    Händelsenamn.
         */
        addObserver (observer, callback, event = "_default") {
            if (!this._callbacks.hasOwnProperty(event)) {
                this._callbacks[event] = [];
            }

            this._callbacks[event].push({
                observer: observer,
                callback: callback
            });
        }

        /**
         * Anropa alla registrerade lyssnarfunktioner för angiven händelse.
         *
         * @param {*}      data  Dataparameter i anropet.
         * @param {string} event Händelsenamn.
         */
        update (data = this, event = "_default") {
            (this._callbacks[event] || []).forEach((entry) => {
                entry.callback.call(entry.observer, data, event);
            });
        }

    }


    /**
     * Ett objekt som observerar andra objekt.
     *
     * Observatören kan begära att bli notifierad när ett observerat objekts
     * värde förändras.
     *
     * @see Observable
     *
     * @method observe Observera ett objekt.
     */
    class Observer {

        /**
         * Registrera en lyssarfunktion hos ett observerbart objekt.
         *
         * @param {Observable} observable Det observerbara objektet.
         * @param {callback}   callback   Lyssnarfunktion.
         * @param {string}     event      Eventuellt händelsenamn.
         */
        observe (observable, callback, event = undefined) {
            if (!(observable instanceof Observable)) {
                console.log(observable);
                throw new Error(
                    "'observable' argument must be instance of Observable."
                );
            }

            observable.addObserver(this, callback, event);
        }

    }


    /**
     * Kryssruta.
     *
     * Anropar observatörernas händelselyssnare när användaren markerar eller
     * avmarkerar kryssrutan.
     *
     * @see Observable
     * @see Observer
     * @property {boolean} checked  Sann om kryssrutan är markerad.
     * @property {boolean} disabled Sann om kryssrutan är inaktiverad.
     */
    class Checkbox extends compound(Observable, Observer) {

        /**
         * Konstruera en kryssruteinstans.
         *
         * @param {HTMLInputElement} cb Kryssruta.
         * @throws {Error}
         */
        constructor (cb) {
            if (!(cb instanceof HTMLInputElement) || "checkbox" !== cb.type) {
                console.log(cb);
                throw new Error(
                    "'cb' argument must be an INPUT element of type checkbox."
                );
            }

            super();

            this._cb = cb;
            this._cb.addEventListener("input", () => this.update());
        }

        // Åtkomstmetoder.
        get checked  ()     {return this._cb.checked }
        get disabled ()     {return this._cb.disabled}

        set disabled (flag) {this._cb.disabled = !!flag;}

    }


    /**
     * Inmatningsfält för färgvärde.
     *
     * Anropar observatörernas händelselyssnare när användaren redigerar värdet
     * i inmatningsfältet.
     *
     * Tolkar färgspecifikationenn i fältet (via en `ColourParser`) och lagrar
     * den som ett RGBA-värde.
     *
     * @see Observable
     * @see Observer
     * @property {string}  colour Värdet i fältet.
     * @property {Object}  rgb    Det tolkade RGBA-värdet.
     * @property {boolean} valid  Anger om specifikationen är giltig.
     */
    class ColourField extends compound(Observable, Observer) {

        /**
         * Konstruera en fältinstans.
         *
         * Lägg till en händelselyssnare för fältet, som anropar alla
         * observatörers registrerade händelselyssnare när användaren ändrar
         * fältets värde.
         *
         * Lägg till en händelselyssnare för fältet, som visar ett felmeddelande
         * om värdet är ogiltigt (och inte tomt) när fältet förlorar fokus.
         *
         * @param {HTMLInputField} field  Inmatningsfält.
         * @param {ColourParser}   parser Färgspecifikationstolk.
         * @throws {Error}
         */
        constructor (field, parser) {
            super();

            this._field  = field;
            this._parser = parser;
            this._rgb    = null;
            this._valid  = false;

            this._field.addEventListener("blur",  () => this.validate());
            this._field.addEventListener("input", () => this.update());
        }

        // Åtkomstmetoder.
        get colour () {return this._field.value}
        get rgb    () {return this._rgb        }
        get valid  () {return this._valid      }

        set colour (colour) {
            this._field.value = colour;
            this.update();
        }

        /**
         * @inheritDoc
         */
        update () {
            try {
                this._rgb   = this._parser.parse(this._field.value);
                this._valid = true;
                this._field.setCustomValidity("");
                this._field.removeAttribute("aria-invalid");

            } catch (err) {
                this._rgb   = null;
                this._valid = false;
                this._field.setCustomValidity(err.message);
                this._field.setAttribute("aria-invalid", "true");
            }

            super.update();
        }

        /**
         * Validera värdet i inmatningsfältet.
         *
         * Om det inte är giltigt, visa ett felmeddelande. Undantag för om
         * fältet är tomt; då ska man kunna lämna fältet ändå.
         */
        validate () {
            this._valid = this._field.checkValidity();

            if (!this._valid && !this._field.validity.valueMissing) {
                this._field.reportValidity();
            }
        }

    }


    /**
     * Tolk för färgspecifikationer.
     *
     * Stödjer följande format:
     * - Hexadecimalt: #3c9, #32C89E
     * - RGB:          rgb(255, 240, 128)
     * - RGBA:         rgba(255, 240, 128, 0.8)
     * - HSL:          hsl(310, 47%, 72%)
     * - HSLA:         hsla(310, 47%, 72%, 0.6)
     * - Nyckelord:    white, DarkSlateBlue
     */
    class ColourParser {

        /**
         * Konvertera HSLA till RGBA.
         *
         * @param {Object} hsla HSLA-värde.
         * @return {Object}
         */
        hslaToRgba (hsla) {
            const hueToRgb = (m1, m2, h) => {
                if (h < 0) h += 1;
                if (h > 1) h -= 1;
                if (h * 6 < 1) return m1 + (m2 - m1) * h * 6;
                if (h * 2 < 1) return m2;
                if (h * 3 < 2) return m1 + (m2 - m1) * (2/3 - h) * 6;

                return m1;
            }

            const h = hsla.h / 360;
            const s = hsla.s / 100;
            const l = hsla.l / 100;

            const m2 = l <= 0.5 ? l * (s + 1) : l + s - l * s;
            const m1 = l * 2 - m2;

            return {
                r: Math.round(255 * hueToRgb(m1, m2, h + 1/3)),
                g: Math.round(255 * hueToRgb(m1, m2, h)),
                b: Math.round(255 * hueToRgb(m1, m2, h - 1/3)),
                a: hsla.a
            };
        }

        /**
         * Tolkar en färgspecifikation.
         *
         * @param string Färgspecifikation att tolka.
         *
         * @return {Object}
         * @throws {Error}
         */
        parse (value) {
            const map = [
                {pattern: /^#/,        method: this.parseHex    },
                {pattern: /^rgba/i,    method: this.parseRgba   },
                {pattern: /^rgb/i,     method: this.parseRgb    },
                {pattern: /^hsla/i,    method: this.parseHsla   },
                {pattern: /^hsl/i,     method: this.parseHsl    },
                {pattern: /^[a-z]+$/i, method: this.parseKeyword}
            ];

            const mapping = map
            .find((mapping) => mapping.pattern.test(value), this);

            if (!mapping) {
                throw new Error(
                    `Fyll i ett färgvärde som hex, rgb(a), hsl(a) eller
CSS-nyckelord.`
                );
            }

            return mapping.method.call(this, value);
        }

        /**
         * Tolka 3- eller 6-ställigt hexadecimalvärde till RGBA.
         *
         * Exempel: #3c9
         * Exempel: #32C89E
         *
         * @param {string} value Hexadecimalvärde att tolka.
         * @return {Object}
         * @throws {Error}
         */
        parseHex (value) {
            const matches = /^#(?:([0-9a-f]{3})|([0-9a-f]{6}))$/i.exec(value);

            if (null !== matches) {
                if (matches[1]) {
                    return {
                        r: parseInt(matches[1][0] + matches[1][0], 16),
                        g: parseInt(matches[1][1] + matches[1][1], 16),
                        b: parseInt(matches[1][2] + matches[1][2], 16),
                        a: 1
                    };

                } else {
                    return {
                        r: parseInt(matches[2].substring(0, 2), 16),
                        g: parseInt(matches[2].substring(2, 4), 16),
                        b: parseInt(matches[2].substring(4, 6), 16),
                        a: 1
                    };
                }
            }

            throw new Error(
                "Fyll i hexvärde med 3 eller 6 tecken, t.ex. #3c9, #32C89E."
            );
        }

        /**
         * Tolka hsl()-notation till RGBA.
         *
         * Exempel: hsl(310, 47%, 72%)
         *
         * @param {string} value Notation att tolka.
         * @return {Object}
         * @throws {Error}
         */
        parseHsl (value) {
            const matches = /^hsl\((\d{1,3}),\s*(\d{1,3})%,\s*(\d{1,3})%\)$/i
                            .exec(value);

            if (null !== matches) {
                return this.hslaToRgba({
                    h: parseInt(matches[1], 10) % 360,
                    s: parseInt(matches[2], 10) % 100,
                    l: parseInt(matches[3], 10) % 100,
                    a: 1
                });
            }

            throw new Error("Fyll i som t.ex. hsl(310, 47%, 72%).");
        }

        /**
         * Tolka hsla()-notation till RGBA.
         *
         * Exempel: hsl(310, 47%, 72%, 0.6)
         *
         * @param {string} value Notation att tolka.
         * @return {Object}
         * @throws {Error}
         */
        parseHsla (value) {
            const matches = /^hsla\((\d{1,3}),\s*(\d{1,3})%,\s*(\d{1,3})%,\s*([01]?(?:\.\d+)?)\)$/i
                            .exec(value);

            if (null !== matches) {
                return this.hslaToRgba({
                    h: parseInt(matches[1], 10) % 360,
                    s: parseInt(matches[2], 10) % 100,
                    l: parseInt(matches[3], 10) % 100,
                    a: parseFloat(matches[4])
                });
            }

            throw new Error("Fyll i som t.ex. hsla(310, 47%, 72%, 0.6).");
        }

        /**
         * Tolka CSS-nyckelord till RGBA.
         *
         * Exempel: white, DarkSlateBlue
         *
         * @param {string} value Nyckelord att tolka.
         * @return {Object}
         * @throws {Error}
         */
        parseKeyword (value) {
            const keywords = {
                aliceblue:            [240, 248, 255],
                antiquewhite:         [250, 235, 215],
                aqua:                 [  0, 255, 255],
                aquamarine:           [127, 255, 212],
                azure:                [240, 255, 255],
                beige:                [245, 245, 220],
                bisque:               [255, 228, 196],
                black:                [  0,   0,   0],
                blanchedalmond:       [255, 235, 205],
                blue:                 [  0,   0, 255],
                blueviolet:           [138,  43, 226],
                brown:                [165,  42,  42],
                burlywood:            [222, 184, 135],
                cadetblue:            [ 95, 158, 160],
                chartreuse:           [127, 255,   0],
                chocolate:            [210, 105,  30],
                coral:                [255, 127,  80],
                cornflowerblue:       [100, 149, 237],
                cornsilk:             [255, 248, 220],
                crimson:              [220,  20,  60],
                cyan:                 [  0, 255, 255],
                darkblue:             [  0,   0, 139],
                darkcyan:             [  0, 139, 139],
                darkgoldenrod:        [184, 134,  11],
                darkgray:             [169, 169, 169],
                darkgreen:            [  0, 100,   0],
                darkgrey:             [169, 169, 169],
                darkkhaki:            [189, 183, 107],
                darkmagenta:          [139,   0, 139],
                darkolivegreen:       [ 85, 107,  47],
                darkorange:           [255, 140,   0],
                darkorchid:           [153,  50, 204],
                darkred:              [139,   0,   0],
                darksalmon:           [233, 150, 122],
                darkseagreen:         [143, 188, 143],
                darkslateblue:        [ 72,  61, 139],
                darkslategray:        [ 47,  79,  79],
                darkslategrey:        [ 47,  79,  79],
                darkturquoise:        [  0, 206, 209],
                darkviolet:           [148,   0, 211],
                deeppink:             [255,  20, 147],
                deepskyblue:          [  0, 191, 255],
                dimgray:              [105, 105, 105],
                dimgrey:              [105, 105, 105],
                dodgerblue:           [ 30, 144, 255],
                firebrick:            [178,  34,  34],
                floralwhite:          [255, 250, 240],
                forestgreen:          [ 34, 139,  34],
                fuchsia:              [255,   0, 255],
                gainsboro:            [220, 220, 220],
                ghostwhite:           [248, 248, 255],
                gold:                 [255, 215,   0],
                goldenrod:            [218, 165,  32],
                gray:                 [128, 128, 128],
                green:                [  0, 128,   0],
                greenyellow:          [173, 255, 47],
                grey:                 [128, 128, 128],
                honeydew:             [240, 255, 240],
                hotpink:              [255, 105, 180],
                indianred:            [205,  92,  92],
                indigo:               [ 75,   0, 130],
                ivory:                [255, 255, 240],
                khaki:                [240, 230, 140],
                lavender:             [230, 230, 250],
                lavenderblush:        [2550, 240, 245],
                lawngreen:            [124, 252,   0],
                lemonchiffon:         [255, 250, 205],
                lightblue:            [173, 216, 230],
                lightcoral:           [240, 128, 128],
                lightcyan:            [224, 255, 255],
                lightgoldenrodyellow: [250, 250, 210],
                lightgray:            [211, 211, 211],
                lightgreen:           [144, 238, 144],
                lightgrey:            [211, 211, 211],
                lightpink:            [255, 182, 193],
                lightsalmon:          [255, 160, 122],
                lightseagreen:        [ 32, 178, 170],
                lightskyblue:         [135, 206, 250],
                lightslategray:       [119, 136, 153],
                lightslategrey:       [119, 136, 153],
                lightsteelblue:       [176, 196, 222],
                lightyellow:          [255, 255, 224],
                lime:                 [  0, 255,   0],
                limegreen:            [ 50, 205,  50],
                linen:                [250, 240, 230],
                magenta:              [255,   0, 255],
                maroon:               [120,   0,   0],
                mediumaquamarine:     [102, 205, 170],
                mediumblue:           [  0,   0, 205],
                mediumorchid:         [186,  85, 211],
                mediumpurple:         [147, 112, 219],
                mediumseagreen:       [ 60, 179, 113],
                mediumslateblue:      [123, 104, 238],
                mediumspringgreen:    [  0, 250, 154],
                mediumturquoise:      [ 72, 209, 204],
                mediumvioletred:      [199,  21, 133],
                midnightblue:         [ 25,  25, 112],
                mintcream:            [245, 255, 250],
                mistyrose:            [255, 228, 225],
                moccasin:             [255, 228, 181],
                navajowhite:          [255, 222, 173],
                navy:                 [  0,   0, 128],
                oldlace:              [253, 245, 230],
                olive:                [128, 128,   0],
                olivedrab:            [107, 142,  35],
                orange:               [255, 165,   0],
                orangered:            [255,  69,   0],
                orchid:               [218, 112, 214],
                palegoldenrod:        [238, 232, 170],
                palegreen:            [152, 251, 152],
                paleturquoise:        [175, 238, 238],
                palevioletred:        [219, 112, 147],
                papayawhip:           [255, 239, 213],
                peachpuff:            [255, 218, 185],
                peru:                 [205, 133, 63],
                pink:                 [255, 192, 203],
                plum:                 [221, 160, 221],
                powderblue:           [176, 224, 230],
                purple:               [120,   0, 128],
                red:                  [255,   0,   0],
                rosybrown:            [188, 143, 143],
                royalblue:            [ 65, 105, 255],
                saddlebrown:          [139,  69,  19],
                salmon:               [250, 128, 114],
                sandybrown:           [244, 164,  96],
                seagreen:             [ 46, 139,  87],
                seashell:             [255, 245, 238],
                sienna:               [160,  82,  45],
                silver:               [192, 192, 192],
                skyblue:              [135, 206, 235],
                slateblue:            [106,  90, 205],
                slategray:            [112, 128, 144],
                slategrey:            [112, 128, 144],
                snow:                 [255, 250, 250],
                springgreen:          [  0, 255, 127],
                steelblue:            [ 70, 130, 180],
                tan:                  [210, 180, 140],
                teal:                 [  0, 128, 128],
                thistle:              [216, 191, 216],
                tomato:               [255,  99,  71],
                turquoise:            [ 64, 224, 208],
                violet:               [238, 130, 238],
                wheat:                [245, 222, 179],
                white:                [255, 255, 255],
                whitesmoke:           [245, 245, 245],
                yellow:               [255, 255,   0],
                yellowgreen:          [154, 205,  50]
            };

            if (keywords.hasOwnProperty(value.toLowerCase())) {
                const [red, green, blue] = keywords[value.toLowerCase()];

                return {
                    r: red,
                    g: green,
                    b: blue,
                    a: 1
                };
            }

            throw new Error("Fyll i ett giltigt CSS-färgnamn.");
        }

        /**
         * Tolka rgb()-notation till RGBA.
         *
         * Exempel: rgb(255, 240, 128)
         *
         * @param {string} value Notation att tolka.
         * @return {Object}
         * @throws {Error}
         */
        parseRgb (value) {
            const matches = /^rgb\((\d{1,3}),\s*(\d{1,3}),\s*(\d{1,3})\)$/i
                            .exec(value);

            if (null !== matches) {
                return {
                    r: parseInt(matches[1], 10) % 256,
                    g: parseInt(matches[2], 10) % 256,
                    b: parseInt(matches[3], 10) % 256,
                    a: 1
                };
            }

            throw new Error("Fyll i som t.ex. rgb(255, 240, 128).");
        }

        /**
         * Tolka rgba()-notation till RGBA.
         *
         * Exempel: rgba(255, 240, 128, 0.8)
         *
         * @param {string} value Notation att tolka.
         * @return {Object}
         * @throws {Error}
         */
        parseRgba (value) {
            const matches = /^rgba\((\d{1,3}),\s*(\d{1,3}),\s*(\d{1,3}),\s*([01]?(?:\.\d+)?)\)$/i
                            .exec(value);

            if (null !== matches) {
                return {
                    r: parseInt(matches[1], 10) % 256,
                    g: parseInt(matches[2], 10) % 256,
                    b: parseInt(matches[3], 10) % 256,
                    a: parseFloat(matches[4])
                };
            }

            throw new Error("Fyll i som t.ex. rgba(255, 240, 128, 0.8).");
        }

    }


    /**
     * Lista med sparade färgvärden.
     *
     * Låter användaren spara och återkalla (eventuellt namngivna) färgvärden.
     * Färgerna sparas persistent, med användarens medgivande.
     *
     * @see Observable
     * @see Observer
     * @property {string} background Aktuell bakgrundsfärg.
     * @property {string} foreground Aktuell förgrundsfärg.
     */
    class ColourStore extends compound(Observable, Observer) {

        /**
         * Konstruera en lagringsinstans.
         *
         * Lägg till händelselyssnare för namnfältet och knapparna. Populera
         * listan med värden som sparats i persistenslagret.
         *
         * @param {HTMLUListElement}  list        Listelement (UL).
         * @param {HTMLInputElement}  nameField   Inmatningsfält för färgnamn.
         * @param {HTMLButtonElement} fgButton    Knapp för att lägga till
         *                                        förgrundsfärg.
         * @param {HTMLButtonElement} bgButton    Knapp för att lägga till
         *                                        bakgrundsfärg.
         * @param {HTMLButtonElement} resetButton Knapp för att rensa sparade
         *                                        färger.
         * @param {PersistentStorage} storage     Persistens.
         * @throws {Error}
         */
        constructor (list, nameField, fgButton, bgButton, resetButton, storage) {
            if (!(list instanceof HTMLUListElement)) {
                console.log(list);
                throw new Error("'list' argument must be a UL element.");
            }

            if (!(nameField instanceof HTMLInputElement)) {
                console.log(nameField);
                throw new Error(
                    "'nameField' argument must be an INPUT element."
                );
            }

            if (!(fgButton instanceof HTMLButtonElement)) {
                console.log(fgButton);
                throw new Error(
                    "'fgButton' argument must be a BUTTON element."
                );
            }

            if (!(bgButton instanceof HTMLButtonElement)) {
                console.log(bgButton);
                throw new Error(
                    "'bgButton' argument must be a BUTTON element."
                );
            }

            if (!(resetButton instanceof HTMLButtonElement)) {
                console.log(resetButton);
                throw new Error(
                    "'resetButton' argument must be a BUTTON element."
                );
            }

            super();

            this._bgButton    = bgButton;
            this._bgColour    = "";
            this._bgListeners = [];
            this._colours     = [];
            this._fgButton    = fgButton;
            this._fgColour    = "";
            this._fgListeners = [];
            this._list        = list;
            this._nameField   = nameField
            this._resetButton = resetButton;
            this._storage     = storage;

            // Händelselyssnare för namnfältet.
            this._nameField.addEventListener("blur", (event) => {
                if (!this._nameField.checkValidity()) {
                    this._nameField.reportValidity();
                }
            });

            this._nameField.addEventListener("input", (event) => {
                if (-1 === this._nameField.value.indexOf("|")) {
                    this._nameField.setCustomValidity("");
                    this._nameField.removeAttribute("aria-invalid");
                } else {
                    this._nameField.setCustomValidity(
                        "Namnet får inte innehålla '|'."
                    );
                    this._nameField.setAttribute("aria-invalid", true);
                }
            });

            // Händelselyssnare för knapparna.
            this._fgButton.addEventListener("click", (event) => {
                event.stopPropagation();
                event.preventDefault();
                this.add(this._fgColour);
            });

            this._bgButton.addEventListener("click", (event) => {
                event.stopPropagation();
                event.preventDefault();
                this.add(this._bgColour);
            });

            this._resetButton.addEventListener("click", (event) => {
                event.stopPropagation();
                event.preventDefault();
                this.clear();
            });


            // Populera listan med eventuella sparade färger.
            this._colours = this._storage.read();
            this._colours.forEach((colour) => this.addListItem(colour));

            this._resetButton.disabled = 0 === this._colours.length;
        }

        // Åtkomstmetoder.
        set background (field) {
            this._bgColour = field.valid ? field.colour : "";
            this.enableButtons();
        }

        set foreground (field) {
            this._fgColour = field.valid ? field.colour : "";
            this.enableButtons();
        }

        set name (name) {this._name = name;}

        /**
         * Lägg till en färg  i listan med sparade färger.
         *
         * Om det är tillåtet, spara färgen i en cookie.
         *
         * @param {string} value Färgvärde att spara.
         */
        add (value) {
            const colour = new NamedColour(this._nameField.value, value);

            this._colours.push(colour);
            this.addListItem(colour);

            this.save();

            // Rensa namnfältet och aktivera rensaknappen.
            this._nameField.value      = "";
            this._resetButton.disabled = false;
        }

        /**
         * Lägg till ett LI-element i UL-listan.
         *
         * @param {NamedColour} colour Färginformation.
         */
        addListItem (colour) {
            // LI-element med kantlinje i angiven färg.
            const li = this._list.appendChild(doc.createElement("li"));
            li.style.borderColor = colour.colour;

            // SPAN-element med eventuellt angiven beteckning.
            li.appendChild(doc.createElement("span")).
                appendChild(doc.createTextNode(colour.label));

            // SPAN-element med färgspecifikationen.
            li.appendChild(doc.createElement("span")).
                appendChild(doc.createTextNode(colour.colour));

            // Knapp för att använda färgen som förgrundsfärg i analysatorn.
            const fg = li.appendChild(doc.createElement("button"));
            fg.type  = "button";
            fg.value = colour.colour;
            fg.title = `Använd ${colour.label} som förgrundsfärg`;
            fg.appendChild(doc.createTextNode("Fg"));
            fg.addEventListener("click", (event) => {
                this.update(event.target.value, "foreground");
            });

            // Knapp för att använda färgen som bakgrundsfärg i analysatorn.
            const bg = li.appendChild(doc.createElement("button"));
            bg.type  = "button";
            bg.value = colour.colour;
            bg.title = `Använd ${colour.label} som bakgrundsfärg`;
            bg.appendChild(doc.createTextNode("Bg"));
            bg.addEventListener("click", (event) => {
                this.update(event.target.value, "background");
            });

            // Knapp för att ta bort en rad ur listan.
            const del = li.appendChild(doc.createElement("button"));
            del.type  = "button";
            del.title = "Ta bort färg";
            del.appendChild(doc.createTextNode("\u2718"));
            del.addEventListener("click", (event) => {
                const li = event.target.parentNode;
                const ul = li.parentNode;
                ul.removeChild(li);
                this.remove(event.target.value);
                this._resetButton.disabled = 0 === this._colours.length;
                this.save();
            });
        }

        /**
         * Rensa listan med sparade färger och UL-listan.
         *
         * Rensa även i persistenslagret.
         */
        clear () {
            this._colours               = [];
            this._list.innerHTML        = "";
            this._resetButton.disabled = true;
            this.save();
        }

        /**
         * Aktivera eller inaktivera sparaknapparna.
         *
         * Aktivera eller inaktivera knapparna för att spara förgrunds- och
         * bakgrundsfärg beroende på om motsvarande värde är satt.
         */
        enableButtons () {
            this._fgButton.disabled = "" === this._fgColour;
            this._bgButton.disabled = "" === this._bgColour;
        }

        /**
         * Ta bort ett sparat färgvärde ur listan.
         *
         * @param {number} id Färgidentifierare.
         */
        remove (id) {
            const index = this._colours.map((colour) => colour.id).indexOf(id);
            this._colours.splice(index, 1);
        }

        /**
         * Spara listan med sparade färger i persistenslagret.
         */
        save () {
            this._storage.write(this._colours);
        }

    }


    /**
     * Kontrastanalysator.
     *
     * Beräknar kontrastförhållandet mellan två färger enligt den algoritm
     * som beskrivs i WCAG 2.1.
     *
     * @see https://www.w3.org/TR/WCAG21/#dfn-contrast-ratio
     */
    class ContrastAnalyzer {

        /**
         * Returnera till vilken nivå en kontrast uppfyller kraven.
         *
         * Returnera "a", "aa" eller "aaa" beroende på kontrastvärde och
         * om det avser stor text eller annat än text.
         *
         * @param {number}  constrast Kontrastvärde.
         * @param {boolean} largeText Kontrollera mot kraven för stor text?
         * @param {boolean} nonText   Kontrollera mot kraven för annat än text?
         *
         * @return String
         */
        compliance (contrast, largeText, nonText) {
            if (nonText) {
                return contrast >= 3 ? "aa" : "a";
            } else if (largeText) {
                return contrast >= 4.5 ? "aaa" : (contrast >= 3 ? "aa" : "a");
            }

            return contrast >= 7 ? "aaa" : (contrast >= 4.5 ? "aa" : "a");
        }

        /**
         * Beräkna kontrastförhållande mellan förgrund och bakgrund.
         *
         * @param {Object} bg Bakgrund som RGB.
         * @param {Object} fg Förgrund som RGB.
         * @return {number}
         */
        contrast (bg, fg) {
            const fromSRgb = (s) => s <= .03928
                ? s / 12.92
                : Math.pow((s + .055) / 1.055, 2.4);

            const convertRgb = (rgb) => ({
                r: fromSRgb(rgb.r / 255),
                g: fromSRgb(rgb.g / 255),
                b: fromSRgb(rgb.b / 255)
            });

            const luminance = (rgb) =>
                rgb.r * .2126 + rgb.g * .7152 + rgb.b * .0722;

            const l1 = luminance(convertRgb(bg));
            const l2 = luminance(convertRgb(fg));

            return Math.round(
                (  Math.max(l1, l2) + .05)
                / (Math.min(l1, l2) + .05)
                * 100
            ) / 100;
        }

        /**
         * Normalisera färger.
         *
         * Beräkna den effektiva färgen för en förgrundsfärg med transparens,
         * där även hela elementet kan ha transparens.
         *
         * Returnera ett RGBA-värde.
         *
         * @param {Object} fg      Förgrundsfärg som RGBA.
         * @param {Object} bg      Bakgrundsfärg som RGB.
         * @param {number} opacity Elementets opacitet.
         * @return {Object}
         */
        normalize (fg, bg, opacity) {
            const normalizeValue = (fgVal, bgVal, alpha) =>
                bgVal + Math.round((fgVal - bgVal) * alpha * opacity);

            return {
                r: normalizeValue(fg.r, bg.r, fg.a),
                g: normalizeValue(fg.g, bg.g, fg.a),
                b: normalizeValue(fg.b, bg.b, fg.a),
                a: 1
            };
        }

    }


    /**
     * Dialogruta för kontrastberäkning.
     *
     * Representerar användargränssnittet i verktyget för kontrast-
     * beräkning.
     *
     * @see Observable
     * @see Observer
     * @property {ColourField} background Aktuell bakgrundsfärg.
     * @property {string}      compliance Konformitetsnivå.
     * @property {number}      contrast   Kontrastförhållande.
     * @property {Object}      effective  Effektiv, RGBA.
     * @property {ColourField} foreground Aktuell förgrundsfärg.
     * @property {boolean}     largeText  Testa för stor text.
     * @property {boolean}     nonText    Testa för annat än text.
     * @property {number}      opacity    Aktuell opacitet.
     */
    class ContrastDialog extends compound(Observable, Observer) {

        /**
         * Konstruera en dialoginstans.
         *
         * Omvandla attributen `data-for` och `data-id` i mallkoden till
         * faktiska `for` och `id`.
         *
         * Lägg till händelselyssnare för LEGEND-elemeentet, som är det enda som
         * syns när dialogrutan är minimerad, och för stängknappen.
         *
         * Lägg till händelselyssnare för dokumentet, som fångar upp mus- och
         * tangentbordshändeler som ska påverka dialogrutan.
         *
         * @param {HTMLFormElement}  form     Formulärelement.
         * @param {ContrastAnalyzer} analyzer Kontrastanalysator.
         *
         * @throws Error
         */
        constructor (form, analyzer) {
            if (!(form instanceof HTMLFormElement)) {
                throw new Error("'form' argument must be a FORM element.");
            }

            super();

            this._analyzer   = analyzer;
            this._bgColour   = "";
            this._bgRgb      = {r: 0, g: 0, b: 0, a: 0};
            this._compliance = "";
            this._contrast   = 0;
            this._effective  = {r: 0, g: 0, b: 0, a: 1};
            this._fgColour   = "";
            this._fgRgb      = {r: 0, g: 0, b: 0, a: 1};
            this._form       = form;
            this._focus      = null;
            this._largeText  = false;
            this._legend     = form.querySelector("fieldset>legend");
            this._nonText    = false;
            this._opacity    = 1;

            // Sätt for- och id-attribut.
            form.querySelectorAll("[data-for]").forEach((element) => {
                element.htmlFor = element.dataset.for;
            });

            form.querySelectorAll("[data-id]").forEach((element) => {
                element.id = element.dataset.id;
            });

            // Lägg till händelselyssnare
            this._legend.addEventListener("click", (event) => {
                    if (0 === event.button) {
                        this.show();
                    }
            });

            if ("undefined" !== typeof EyeDropper) {
                const showEyeDropper = (event) => {
                    event.preventDefault();
                    event.stopPropagation();
                    const btn = (e) => e instanceof HTMLButtonElement
                        ? e : e ? btn(e.parentNode) : e;
                    const field = btn(event.target)?.dataset.target;
                    const eyeDropper = new EyeDropper();
                    eyeDropper.open()
                    .then((colour) => {
                        fields[field].colour = colour.sRGBHex;
                    })
                    .catch((err) => {});
                };
                form.querySelectorAll("button.eyedropper").forEach((btn) => {
                    btn.disabled = false;
                    btn.addEventListener("click", showEyeDropper);
                });
            }

            form.querySelector("button.close").
                addEventListener("click", (event) => {
                    this.minimize()
            });


            doc.addEventListener("focus",   (ev) => this.trapFocus(ev), true);
            doc.addEventListener("keyup",   (ev) => this.onKeyUp(ev));
            doc.addEventListener("mouseup", (ev) => this.onMouseUp(ev));
        }

        // Åtkomstmetoder.
        get compliance () {return this._compliance}
        get contrast   () {return this._contrast  }
        get effective  () {return this._effective }

        set background (field) {
            if (field.valid) {
                this._bgColour = field.colour;
                this._bgRgb    = field.rgb;

                this.update();
            }
        }

        set foreground (field) {
            if (field.valid) {
                this._fgColour = field.colour;
                this._fgRgb    = field.rgb;

                this.update();
            }
        }

        set nonText (flag) {
            this._nonText  = flag;
            this.update();
        }

        set largeText (flag) {
            this._largeText = flag;
            this.update();
        }

        set opacity (opacity) {
            this._opacity   = opacity;
            this.update();
        }

        /**
         * Minimera dialogrutan.
         *
         * Ta bort den klass som via CSS gör hela dialogrutan synlig.
         */
        minimize () {
            this._form.classList.remove("open");
            this._focus           = null;
            this._legend.tabIndex = 0;
        }

        /**
         * Hantera tangenttryckningar.
         *
         * Escape döljer dialogrutan.
         * Ctrl+K visar dialogrutan.
         *
         * @param {KeyboardEvent} event Händelseinformation.
         */
        onKeyUp (event) {
            let handled = false;
            if (!event.shiftKey && !event.altKey && !event.metaKey) {
                switch (event.key) {
                    case "Escape":
                        if (!event.ctrlKey) {
                            this.minimize();
                            handled = true;
                        }
                        break;
                    case "k":
                        if (event.ctrlKey) {
                            this.show();
                            handled = true;
                        }
                        break;
                }
                if (handled) {
                    event.preventDefault();
                    event.stopPropagation();
                }
            }
        }

        /**
         * Hantera mushändelser.
         *
         * Om musknapp 1 släpps upp utanför dialogrutan, minimera dialogrutan.
         *
         * @param {MouseEvent} event Händelseinformation.
         */
        onMouseUp (event) {
            if (0 === event.button && !this._form.contains(event.target)) {
                this.minimize();
            }
        }

        /**
         * Visa dialogrutan.
         *
         * Sätt en klass som via CSS gör hela dialogrutan synlig.
         * Sätt fokus till inmatningsfältet för förgrundsfärg.
         */
        show () {
            this._form.classList.add("open");
            this._focus = this._form["ca-foreground"];
            this._focus.focus();
            this._legend.tabIndex = -1;
        }

        /**
         * Håll tangentbordsfokus inom dialogrutan.
         *
         * Håll reda på vilket element i dialogrutan som senast hade fokus. Om
         * användaren försöker navigera ut ur dialogen med tangentbordet från
         * första eller sista elementet, sätt fokus till sista respektive första
         * elementet.
         *
         * @param {FocusEvent} event Händelseinformation.
         */
        trapFocus (event) {
            if (null !== this._focus) {
                const focusable = this._form
                .querySelectorAll("button:not(:disabled), input, summary");

                if (this._form.contains(event.target)) {
                    // Spara senast fokuserade element inom dialogrutan.
                    this._focus = event.target;

                } else if (this._focus === focusable[0]) {
                    // Fokus lämnar första elementet, sätt fokus till sista.
                    focusable[focusable.length - 1].focus();

                } else {
                    // Fokus lämnar sista elementet, sätt fokus till första.
                    focusable[0].focus();
                }
            }
        }

        /**
         * Uppdatera beräkningsresultaten.
         *
         * - Normalisera bakgrundsfärgen mot vitt.
         * - Normalisera förgrundsfärgen mot bakgrundsfärgen.
         * - Beräkna och visa kontrastförhållande.
         * - Åskådliggör genom att åsätta färgerna på exemplet.
         * - Visa effektiv färg (i 6-siffrig hexadecialnotation).
         * - AKtivera sparaknapparna.
         */
        update () {
            const norm = (fg, bg, op) => this._analyzer.normalize(fg, bg, op);

            const white      = {r: 255, g: 255, b: 255, a: 1};
            const bg         = norm(this._bgRgb, white, this._opacity);
            const fg         = norm(this._fgRgb, bg,    this._opacity);
            this._effective  = fg;
            this._contrast   = this._analyzer.contrast(fg, bg);
            this._compliance = this._analyzer.compliance(
                                   this._contrast,
                                   this._largeText,
                                   this._nonText
                               );

            super.update();
        }

    }


    /**
     * Visningsfält för beräknade värden.
     *
     * @see Observer
     * @property {string}              className HTML-elementets class-attribut.
     * @property {string}              content   HTML-elementets interna HTML.
     * @property {CSSStyleDeclaration} style     HTNL-elementets style-egenskap.
     * @property {string}              value     HTML-elementets value-attribut.
     */
    class DisplayField extends Observer {

        /**
         * Konstruera en fältinstans.
         *
         * @param {HTMLOutptuElement} element HTML-element.
         */
        constructor (element) {
            super();
            this._element = element;
        }

        // Åtkomstmetoder.
        get style     ()      {return this._element.style     }
        set className (cls)   {this._element.className = cls  }
        set content   (html)  {this._element.innerHTML = html }
        set value     (value) {this._element.value     = value}

    }


    /**
     * Information om en sparad färgspecifikation.
     *
     * Innehåller en färgspecifikation (sträng) och eventuell beteckning som
     * användaren åsatt.
     *
     * @property {string} colour Färgspecifikation.
     * @property {number} id     Identifierare.
     * @property {string} label  Användarens beteckning.
     */
    class NamedColour {

        // Global räknare för id-värden.
        static id = 0;

        /**
         * Konstruera en instans.
         *
         * @param {string} label  Användarens beteckning.
         * @param {string} colour Färgspecifikation.
         * @param {number} id     Unik identifierare.
         */
        constructor (label, colour) {
            this._label  = label;
            this._colour = colour;
            this._id     = ++NamedColour.id;
        }

        // Åtkomstmetoder.
        get colour () {return this._colour}
        get id     () {return this._id    }
        get label  () {return this._label }
    }


    /**
     * Inmatningsfält för numeriskt värde.
     *
     * @see Observable
     * @property {number} value Fältvärde.
     */
    class NumberField extends Observable {

        /**
         * Konstruera en fältinstans.
         *
         * Lägg till en händelselyssnare för fältet, som anropar alla
         * observatörers registrerade händelselyssnare när användaren ändrar
         * fältets värde.
         *
         * Lägg till en händelselyssnare för fältet, som visar ett felmeddelande
         * om värdet är ogiltigt (och inte tomt) när fältet förlorar fokus.
         *
         * @param {HTMLInputField} field Inmatningsfält.
         * @throws {Error}
         */
        constructor (field) {
            super();

            this._field = field;
            this._valid = false;

            this._field.addEventListener("blur",  () => this.validate());
            this._field.addEventListener("input", () => this.update());
        }

        // Åtkomstmetoder.
        get value () {return this._field.valueAsNumber}

        /**
         * @inheritDoc
         */
        update () {
            this._valid = this._field.checkValidity();

            if (this._valid) {
                super.update();
            }
        }

        /**
         * Validera värdet i inmatningsfältet.
         *
         * Om det inte är giltigt, visa ett felmeddelande. Undantag för om
         * fältet är tomt; då ska man kunna lämna fältet ändå.
         */
        validate () {
            if (!this._valid && !this._field.validity.valueMissing) {
                this._field.reportValidity();
            }
        }

    }


    /**
     * Persistent lagring av färgvärden.
     *
     * Persistenslagret består av en cookie (`manual.X.colours`) per typ av
     * granskningsobjekt (webbsida, dokument, mobil applikation).
     *
     * Klassen sätter inga cookies utan användarens uttryckliga medgivande (som
     * persisteras med cookien `manual.cookies` i `fail.js`.
     *
     * @method expires Returnerar tidsstämpeln då cookien ska upphöra.
     * @method read    Läs och returnera sparade färger i persistenslagret.
     * @method write   Spara färger i persistenslagret.
     */
    class PersistentStorage {

        /**
         * Konstruera en lagringsinstans.
         *
         * @param {string} key  Lagringsnyckel: "web", "doc" eller "app".
         * @param {number} days Lagringstid (dagar).
         */
        constructor (key, days) {
            this._active = /(^|\s*)manual\.cookies=allow(;|$)/.test(doc.cookie);
            this._days   = days;
            this._key    = key;
        }

        /**
         * Returnera utgångstid för cookie.
         *
         * @return {Date}
         */
        expires () {
            const date = new Date();
            date.setTime(date.getTime() + this._days * 24 * 60 * 60 * 1000);

            return date;
        }

        /**
         * Läs sparade färger från persistenslagret.
         *
         * @return {NamedColour[]}
         */
        read () {
            const colours = [];
            const pattern = `(?:^|;\\s*)manual\\.${this._key}\\.colours=([^;]*)(?:;|$)`;
            const matches = RegExp(pattern).exec(doc.cookie);

            if (null !== matches) {
                const pattern = /([^:]*):([^|]+)(?:\||$)/g
                let submatches;
                while (null !== (submatches = pattern.exec(matches[1]))) {
                    colours.push(new NamedColour(
                        decodeURIComponent(submatches[1]),
                        submatches[2]
                    ));
                }
            }

            return colours;
        }

        /**
         * Spara färger i persistenslagret.
         *
         * @param {NamedColour[]} colours Namngivna färgspecifikationer.
         */
        write (colours) {
            if (!this._active) {
                // Användaren har inte samtyckt till att använda cookies.
                return;
            }

            // Om det inte finns några sparade färger, radera cookien.
            const expires = 0 === colours.length
                            ? "Thu, 01 Jan 1970 00:00:01 GMT"
                            : this.expires().toUTCString();

            const value = colours
                .map((c) => `${encodeURIComponent(c.label)}:${c.colour}`)
                .join("|");
            doc.cookie = `manual.${this._key}.colours=${value};expires=${expires};path=/`;
        }
    }


    /*
     * Initiering.
     */

    // Kontrollera kompatibilitet.
    if (   "undefined" === typeof doc.body.classList
        || "undefined" === typeof doc.querySelectorAll
        || "undefined" === typeof HTMLInputElement.prototype.setCustomValidity
    ) {
        return;
    }

    // Klona mallkoden och infoga klonen sist i MAIN.
    const main = doc.querySelector("main");
    main.appendChild(
        doc.getElementById("contrast-analyzer").content.cloneNode(true)
    );

    // Dialogruta för kontrastanalys.
    const field  = (id) => doc.getElementById(id);
    const form   = main.lastElementChild;
    const dialog = new ContrastDialog(form, new ContrastAnalyzer());

    // Färglagringsfunktion.
    const [fgBtn, bgBtn, resetBtn] = form.querySelectorAll(".ca-save button");
    const store = new ColourStore(
        form.querySelector("ul.ca-saved-colours"),
        field("ca-name"),
        fgBtn,
        bgBtn,
        resetBtn,
        new PersistentStorage(doc.body.dataset.persistence, 30)
    );

    // Inmatningsfält.
    const parser     = new ColourParser();
    const foreground = new ColourField(field("ca-foreground"), parser);
    const background = new ColourField(field("ca-background"), parser);
    const opacity    = new NumberField(field("ca-opacity"));
    const largeText  = new Checkbox(field("ca-large"));
    const nonText    = new Checkbox(field("ca-non-text"));
    const fields     = {background, foreground};

    // Observatörer.
    dialog.observe(foreground, (input) => dialog.foreground   = input);
    dialog.observe(background, (input) => dialog.background   = input);
    dialog.observe(opacity,    (input) => dialog.opacity      = input.value);
    dialog.observe(largeText,  (cb)    => dialog.largeText    = cb.checked);
    dialog.observe(nonText,    (cb)    => dialog.nonText      = cb.checked);
    nonText.observe(largeText, (cb)    => nonText.disabled    = cb.checked);
    largeText.observe(nonText, (cb)    => largeText.disabled  = cb.checked);

    foreground.observe(
        store,
        (colour) => foreground.colour = colour,
        "foreground"
    );
    background.observe(
        store,
        (colour) => background.colour = colour,
        "background"
    );
    store.observe(
        foreground,
        (field)  => store.foreground = field
    );
    store.observe(
        background,
        (field)  => store.background = field
    );

    const effective = new DisplayField(field("ca-effective"));
    effective.observe(dialog, (dlg) => {
        const toHex = (rgb) =>
            ("000000"
             + new Number((rgb.r << 16) + (rgb.g << 8) + rgb.b).toString(16))
            .substr(-6);
        effective.value = `#${toHex(dlg.effective)}`;
    });

    const ratio = new DisplayField(field("ca-ratio"));
    ratio.observe(dialog, (dlg) => {
        ratio.content   = String(Math.round(dlg.contrast * 100) / 100)
                          .replace('.', ',') + ":1";
        ratio.className = dlg.compliance;
    });

    const sample = new DisplayField(field("ca-sample"));
    sample.observe(
        foreground,
        (input) => sample.style.color = input.colour
    );
    sample.observe(
        background,
        (input) => sample.style.backgroundColor = input.colour
    );
    sample.observe(
        opacity,
        (input) => sample.style.opacity = String(input.value)
    );

    // Sätt en klass på BODY för att tillämpa CSS-regler.
    doc.body.classList.add("js-ca");
})(document);
